extern "C" {
#include <src/cluster.h>
}
#include "otree.h"

void pmlotree(int nele, int maxn, double ogap, double** distmatrix, vector<Onode> & result)
/*
Purpose
=======

The pmlotree routine performs clustering using pairwise max linkage on
the given distance matrix while allow overlapping

Arguments
=========

nele: # elements to be clustered.

distmatrix: the distance matrix, with nele rows, each row being filled
up to the diagonal. The elements on the diagonal exist. The distance
matrix will be modified by this routine.

result: touse to return vector of Onode structs, describing the
clustering solution consisting of a set of nodes. See src/cluster.h for
a description of the Onode structure.

If a memory error occurs, pmlotree returns NULL???
==================================================================
*/
{
  // curn: # potential merging candidates
  int i, j, curn = nele, is, js, ks, flag;
  double d, selfd, min, mergedtouse;
  Onode curnode;
  // clusterid: each candidate in distmatrix is a cluster (subtree).
  //  Assign it an ID. The singletons will have the ID equal to their
  //  original position in distmatrix. The cluster created through
  //  merging is assigned an ID, which is one less the negative of its
  //  postion in result
  intvec clusterid(maxn), mergedomat(maxn), mergedmmat(maxn);
  // info about 'touse' is also in results except for the singletons,
  //   which can be not equal to g_dlarge. May remove 'touse' if we can
  //   take care of that
  dblvec mergedist(maxn);
  // omat: absolute value: # overlap. neg: one is subset
  // mmat: 1: merged. 2: if merge, creat identical cluster to someone
  //   within ogap so merge skipped. 0: otherwise
  vector<intvec> omat(maxn), mmat(maxn);
  // subclus: all clusters that's a subset of the to-be-merged cluster
  intset newclus, subclus;
  intset::const_iterator si, si1;

  result.clear();
  for (i = 0; i < maxn; i++) {
    omat[i].resize(i + 1);
    mmat[i].resize(i + 1);
    for (j = 0; j < i; j++) {
      omat[i][j] = 0;
      mmat[i][j] = 0;
    }
    omat[i][i] = -1;
    mmat[i][i] = 2;
  }
  // setup a list specifying to which cluster each singleton belongs,
  // & keep track of # elements in each cluster (needed to calculate
  // the average)
  for (i = 0; i < nele; i++) {
    clusterid[i] = i;
    curnode.left = -1;
    curnode.right = -1;
    curnode.distance = 0;
    curnode.upperb = g_dlarge;
    curnode.membs.clear();
    curnode.membs.insert(i);
    result.push_back(curnode);
  }

  while (curn > 1) {
    cout << "new iter with " << curn << " clusters" << endl;
    is = 1;
    js = 0;
    // closest pair not excluded by omat or mmat, but might merge into
    // a cluster that's identical to a previous one
    //
    // d should be non-decreasing, which is the basis of many
    // following asserts.
    d = find_closest_pair_pmlotree(curn, distmatrix, &is, &js, omat, mmat);
    if (d == g_dlarge) {
      // we can't find any close pair
      cout << "can't find any close pair" << endl;
      break;
    }
    Onode & nis = result[clusterid[is]];
    Onode & njs = result[clusterid[js]];
    if (g_debug >= 1) {
      cout << "closest pair:\t" << is << '\t' << js << endl;
      Mylib::PrintJoin(cout, nis.membs, " ");
      cout << endl;
      Mylib::PrintJoin(cout, njs.membs, " ");
      cout << endl << "distance: " << d << endl;
    }
    flag = 0;
    if (d - nis.upperb < ogap && d - njs.upperb < ogap) {
      const intset & iset = nis.membs;
      const intset & jset = njs.membs;
      newclus = iset;
      for (si = jset.begin(); si != jset.end(); si++) {
	newclus.insert(*si);
      }
      if (g_debug) {
	cout << "newclus\t";
	Mylib::PrintJoin(cout, newclus, " ");
	cout << endl;
      }

      selfd = max(d, max(distmatrix[is][is], distmatrix[js][js]));
      mergedtouse = g_dlarge;
      subclus.clear();
      /* Fix the distances */
      for (i = 0; i < js; i++) {
	Onode & ni = result[clusterid[i]];
	const intset & cset = ni.membs;
	mergedomat[i] = CalcOverlap(omat[is][js], omat[is][i], omat[js][i], newclus, iset, jset, cset);
	if (-mergedomat[i] == newclus.size() && -mergedomat[i] == cset.size()) {
	  // newclus is identical to cset
	  assert(false);
	  cout << is << ' ' << js << " merge to " << i << endl;
	  flag = -1;
	  break;
	}
	mergedmmat[i] = 0;
	mergedist[i] = max(selfd, max(distmatrix[i][i], max(distmatrix[is][i], distmatrix[js][i])));
	if (-mergedomat[i] == newclus.size()) {
	  assert(ni.distance == d);
	  assert(false);
	  cout << is << ' ' << js << " merge to subset of " << i << endl;
	  flag = -1;
	  break;
	}
	if (-mergedomat[i] == cset.size()) {
	  subclus.insert(i);
	  mergedmmat[i] = 2;
	  if (ni.upperb > mergedist[i]) {
	    assert(ni.upperb == g_dlarge);
	    ni.upperb = mergedist[i];
	  }
	}
      }
      if (flag != -1) {
	for (i = js; i < is; i++) {
	  Onode & ni = result[clusterid[i]];
	  const intset & cset = ni.membs;
	  mergedomat[i] = CalcOverlap(omat[is][js], omat[is][i], omat[i][js], newclus, iset, jset, cset);
	  if (-mergedomat[i] == newclus.size() && -mergedomat[i] == cset.size()) {
	    // if newclus is identical to cset
	    assert(false);
	    cout << is << ' ' << js << " merge to " << i << endl;
	    flag = -1;
	    break;
	  }
	  mergedmmat[i] = (i == js? 2: 0);
	  mergedist[i] = max(selfd, max(distmatrix[i][i], max(distmatrix[is][i], distmatrix[i][js])));
	  if (-mergedomat[i] == newclus.size()) {
	    assert(ni.distance == d);
	    assert(false);
	    cout << is << ' ' << js << " merge to subset of " << i << endl;
	    flag = -1;
	    break;
	  }
	  if (-mergedomat[i] == cset.size()) {
	    subclus.insert(i);
	    mergedmmat[i] = 2;
	    if (ni.upperb > mergedist[i]) {
	      assert(ni.upperb == g_dlarge);
	      ni.upperb = mergedist[i];
	    }
	  }
	}
      }
      if (flag != -1) {
	for (i = is; i < curn; i++) {
	  Onode & ni = result[clusterid[i]];
	  const intset & cset = ni.membs;
	  mergedomat[i] = CalcOverlap(omat[is][js], omat[i][is], omat[i][js], newclus, iset, jset, cset);
	  if (-mergedomat[i] == newclus.size() && -mergedomat[i] == cset.size()) {
	    // if newclus is identical to cset
	    assert(false);
	    cout << is << ' ' << js << " merge to " << i << endl;
	    flag = -1;
	    break;
	  }
	  mergedmmat[i] = (i == is? 2: 0);
	  mergedist[i] = max(selfd, max(distmatrix[i][i], max(distmatrix[i][is],distmatrix[i][js])));
	  if (-mergedomat[i] == newclus.size()) {
	    assert(ni.distance == d);
	    assert(false);
	    cout << is << ' ' << js << " merge to subset of " << i << endl;
	    flag = -1;
	    break;
	  }
	  if (-mergedomat[i] == cset.size()) {
	    subclus.insert(i);
	    mergedmmat[i] = 2;
	    if (ni.upperb > mergedist[i]) {
	      assert(ni.upperb == g_dlarge);
	      ni.upperb = mergedist[i];
	    }
	  }
	}
      }
      // do following regardless of flag
      if (nis.upperb > d) {
	assert(nis.upperb == g_dlarge);
	nis.upperb = d;
      }
      if (njs.upperb > d) {
	assert(njs.upperb == g_dlarge);
	njs.upperb = d;
      }
      if (flag == -1) {
	mmat[is][js] = 2;
      } else {
	flag = 1;
	mmat[is][js] = 1;
	// cout << "mergedist" << endl;
	// Mylib::PrintJoin(cout, mergedist, " ");
	// cout << endl;
	// cout << "selfd\t" << selfd << endl;
	// cout << "mergedomat" << endl;
	// Mylib::PrintJoin(cout, mergedomat, " ");
	// cout << endl;
	// cout << "mergedmmat" << endl;
	// Mylib::PrintJoin(cout, mergedmmat, " ");
	// cout << endl;
	/* Save result */
	curnode.distance = d;
	curnode.upperb = mergedtouse;
	curnode.left = clusterid[is];
	curnode.right = clusterid[js];
	curnode.membs = newclus;
	result.push_back(curnode);
	if (result[curnode.left].upperb > d) {
	  result[curnode.left].upperb = d;
	}
	if (result[curnode.right].upperb > d) {
	  result[curnode.right].upperb = d;
	}
	// may optimize the following
	if (g_debug) {
	  cout << "subclus\t" << subclus.size() << endl;
	  Mylib::PrintJoin(cout, subclus, " ");
	  cout << endl;
	}
	for (si = subclus.begin(); si != subclus.end(); si++) {
	  for (si1 = subclus.begin(); si1 != subclus.end(); si1++) {
	    if (*si > *si1) {
	      mmat[*si][*si1] = 2;
	    }
	  }
	}
      }
    }
    if (g_debug) {
      cout << "flag\t" << flag << endl;
    }
    // PrintVecmat(cout, mmat, curn);
    for (i = curn - 1; i >= 0; i--) {
      if (d - result[clusterid[i]].upperb >= ogap) {
	if (g_debug) {
	  cout << "remove cluster " << i << " from the pool:" << endl;
	  Mylib::PrintJoin(cout, result[clusterid[i]].membs, " ");
	  cout << endl;
	}
	curn--;
	if (i < curn) {
	  for (j = 0; j < i; j++) {
	    distmatrix[i][j] = distmatrix[curn][j];
	    omat[i][j] = omat[curn][j];
	    mmat[i][j] = mmat[curn][j];
	  }
	  distmatrix[i][i] = distmatrix[curn][curn];
	  omat[i][i] = omat[curn][curn];
	  mmat[i][i] = mmat[curn][curn];
	  for (j = i + 1; j <= curn; j++) {
	    distmatrix[j][i] = distmatrix[curn][j];
	    omat[j][i] = omat[curn][j];
	    mmat[j][i] = mmat[curn][j];
	  }
	  clusterid[i] = clusterid[curn];
	  mergedist[i] = mergedist[curn];
	  mergedomat[i] = mergedomat[curn];
	  mergedmmat[i] = mergedmmat[curn];
	}
      }
    }
    // PrintVecmat(cout, omat, curn);
    // PrintVecmat(cout, mmat, curn);
    if (flag == 1) {
      if (curn < maxn) {
	ks = curn;
	curn++;
      } else {
	min = g_dlarge;
	for (i = 0; i < curn; i++) {
	  if (min > result[clusterid[i]].upperb) {
	    min = result[clusterid[i]].upperb;
	    ks = i;
	  }
	}
      }
      for (i = 0; i < ks; i++) {
	distmatrix[ks][i] = mergedist[i];
	omat[ks][i] = mergedomat[i];
	mmat[ks][i] = mergedmmat[i];
      }
      distmatrix[ks][ks] = selfd;
      omat[ks][ks] = -newclus.size();
      mmat[ks][ks] = 2;
      for (i = ks + 1; i < curn; i++) {
	distmatrix[i][ks] = mergedist[i];
	omat[i][ks] = mergedomat[i];
	mmat[i][ks] = mergedmmat[i];
      }
      /* Update clusterids */
      clusterid[ks] = result.size() - 1;
    }
    // cout << "omat" << endl;
    // PrintVecmat(cout, omat, curn);
    // cout << "mmat" << endl;
    // PrintVecmat(cout, mmat, curn);
    // cout << "clusterid" << endl;
    // Mylib::PrintJoin(cout, clusterid, " ");
    // cout << endl;
    // PrintDistmatCluster(cout, distmatrix, curn, result, clusterid);
    // PrintCluster(cout, curn, result, clusterid);
  }
}

/* ******************************************************************** */

void cutotree(double cutgap, const vector<Onode> & otree, double thresh, vector<pair<int, intset> > & clusters)

/*
Purpose
=======

The cutotree routine takes the output of a hierarchical clustering routine, and
divides the elements in the otree structure into clusters based on the
hierarchical clustering result. The number of clusters is specified by the user.

Arguments
=========

nele      (input) int
The number of elements that were clustered.

otree           (input) Onode[nele-1]
The clustering solution. Each node in the array describes one linking event,
with otree[i].left and otree[i].right representig the elements that were joined.
The original elements are numbered 0..nele-1, nodes are numbered
-1..-(nele-1).

nclusters      (input) int
The number of clusters to be formed.

clusterid      (output) int[nele]
The number of the cluster to which each element was assigned. Space for this
array should be allocated before calling the cutotree routine. If a memory
error occured, all elements in clusterid are set to -1.

========================================================================
*/
{
  int i;
  intset membs;
  clusters.clear();
  // not output singletons because we didn't record when it's touse
  for (i = 0; i < otree.size(); i++) {
    if (otree[i].distance <= thresh && thresh - otree[i].upperb < cutgap) {
      membs.clear();
      AddClus(i, otree, membs);
      clusters.push_back(make_pair(i, membs));
    }
  }
}

// The palotree routine performs clustering using pairwise average
// linkage on the given distance matrix while allow overlapping
//
// nele: # elements to be clustered.
//
// maxn: max # candidate clusters ever allowed in pool
//
// dmat: the distance matrix, with nele rows, each row being filled up
// to the diagonal. The elements on the diagonal do not exist. The
// distance matrix will be modified by this routine.
//
// result: used to return vector of Onode structs, describing the
// clusters ever appeared. See util.h for a description of the Onode
// structure.
//
// dentout: output dendogram
//
// opt: 0: exclude subset. 1: exclude exact match
void palotree_tumat(int nele, int maxn, double ogap, double** dmat, vector<Onode> & result, ostream & dendout, int opt)
{
  // curn: # potential merging candidates
  int i, j, curn = nele, is, js, k, flag, iter = 0;
  double prevd, d = -g_dlarge, min, sum, mwi, szsub, mgtouse, di, dj;
  Onode curnode;
  // clusterid: each candidate in dmat is a cluster (subtree).
  //  Assign it an ID. The singletons will have the ID equal to their
  //  original position in dmat. The cluster created through
  //  merging is assigned an ID, which is one less the negative of its
  //  postion in result
  intvec clusterid(maxn), number(maxn), numsub(maxn);
  dblvec within(maxn), touse(maxn);
  // mmat: 1: merged. 2: if merge, creat identical cluster to someone
  // within ogap so merge skipped. 0: otherwise
  vector<intvec> omat(maxn), mmat(maxn);
  vector<dblvec> tumat(maxn), cmat(maxn); // dmat
  //
  intvec mergedomat(maxn), mergedmmat(maxn);
  // info about 'touse' is also in results except for the singletons,
  // which can be not equal to g_dlarge. May remove 'touse' if we can
  // take care of that
  //
  // cmat[merge][i] = mergedcmat1[i]. cmat[i][merge] = mergedcmat2[i]
  // same for tumat
  dblvec mergedmat(maxn), mergedtumat1(maxn), mergedtumat2(maxn), mergedcmat1(maxn), mergedcmat2(maxn);
  // subclus: all clusters that's a subset of the to-be-merged cluster
  intset subclus, supclus, nonol, tupdate;
  intset::const_iterator si, si1;

  // --------------------------------------------------
  //
  // initialize mat's, vec's, and result for singletons
  //
  // --------------------------------------------------
  result.clear();
  curnode.membs.clear();
  for (i = 0; i < maxn; i++) {
    omat[i].resize(i);
    mmat[i].resize(i);
    tumat[i].resize(maxn);
    cmat[i].resize(maxn);
    for (j = 0; j < i; j++) {
      omat[i][j] = 0;
      mmat[i][j] = 0;
      tumat[i][j] = g_dlarge;
      cmat[i][j] = g_dlarge;
    }
    tumat[i][i] = -g_dlarge;
    cmat[i][j] = g_dlarge;
    for (j = i + 1; j < maxn; j++) {
      tumat[i][j] = g_dlarge;
      cmat[i][j] = g_dlarge;
    }
  }
  for (i = 0; i < nele; i++) {
    clusterid[i] = i;
    number[i] = 1;
    numsub[i] = 0;
    within[i] = 0;
    touse[i] = g_dlarge;
    curnode.left = -1;
    curnode.right = -1;
    curnode.distance = 0;
    curnode.upperb = g_dlarge;
    curnode.lowerb = g_dlarge;
    curnode.size = 1;
    result.push_back(curnode);
  }

  while (curn > 1 && iter < g_maxiter) {
    is = 1;
    js = 0;
    // TODO: check if d is non-decreasing
    prevd = d;
    d = find_closest_pair_palotree(curn, dmat, &is, &js, omat, mmat);
    tupdate.clear();
    tupdate.insert(is);
    touse[is] = -g_dlarge;
    tupdate.insert(js);
    touse[js] = -g_dlarge;
    cout << "new iter\t" << iter << "\t# cluster\t" << curn << "\tdistance\t" << d << endl;
    if (d == g_dlarge) {
      cout << "can't find any close pair" << endl;
      break;
    }
    if (g_debug >= 1) {
      cout << "closest pair:\t" << is << '\t' << js << '\t' << number[is] << '\t' << number[js] << endl;
    }
    // --------------------------------------------------
    //
    // flag status of the pair found: merge to existing cluster (< 0),
    // used (0), or okay to merge (1). Fill mergedomat along the way
    // because it's needed for checking the status
    //
    // --------------------------------------------------
    flag = 0;
    if (d - tumat[is][js] < ogap && d - tumat[js][is] < ogap) {
      sum = number[is] + number[js]; // b/c no overlap between is and js
      if (g_debug >= 1) {
	cout << "newclus size\t" << sum << endl;
      }
      for (k = 0; k < curn; k++) {
	if (k == is || k == js) {
	  continue;
	}
	i = (is > k? omat[is][k]: omat[k][is]);
	j = (js > k? omat[js][k]: omat[k][js]);
	mergedomat[k] = i + j;
	if (opt == 0) {
	  // Can remove the following if it's never true
	  if (mergedomat[k] == sum) {
	    cout << is << ' ' << js << " merge to subset of " << k << endl;
	    flag = -k - 1;
	    break;
	  }
	} else {
	  if (mergedomat[k] == sum && mergedomat[k] == number[k]) {
	    cout << is << ' ' << js << " merge to " << k << endl;
	    flag = -k - 1;
	    break;
	  }
	}
      }
      if (flag == 0) {
	flag = 1;
	mergedomat[is] = number[is];
	mergedomat[js] = number[js];
      }
    }

    if (g_debug >= 1) {
      cout << "flag\t" << flag << endl;
    }
    if (flag == 0) {
      mmat[is][js] = 3;
    } else if (flag < 0) {
      mmat[is][js] = 2;
    } else {
      // --------------------------------------------------
      //
      // If okay to merge, fill mergedmat. Minor update to mmat
      //
      // --------------------------------------------------
      curnode.distance = d;
      curnode.upperb = g_dlarge; // will be updated later
      curnode.lowerb = g_dlarge;
      curnode.left = clusterid[is];
      curnode.right = clusterid[js];
      curnode.size = sum;
      result.push_back(curnode);
      mmat[is][js] = 1;
      mwi = within[is] + within[js] + d * number[is] * number[js];
      for (k = 0; k < curn; k++) {
	if (k == is || k == js) {
	  continue;
	}
	di = (is > k? dmat[is][k]: dmat[k][is]);
	dj = (js > k? dmat[js][k]: dmat[k][js]);
	if (mergedomat[k] == 0) {
	  mergedmat[k] = (di * number[is] + dj * number[js]) / sum;
	}
      }
      // --------------------------------------------------
      //
      // fill mergemmat and mergedcmat. Update newsub
      //
      // --------------------------------------------------
      supclus.clear();
      subclus.clear();
      nonol.clear();
      for (k = 0; k < curn; k++) {
	if (mergedomat[k] == sum) { // merged is subset of k
	  assert(opt != 0);
	  mergedmmat[k] = 2;
	  mergedcmat1[k] = (within[k] - mwi) / (number[k] * (number[k] - 1) / 2 - sum * (sum - 1) / 2);
	  mergedcmat2[k] = g_dlarge;
	  supclus.insert(k);
	  numsub[k]++;
	} else if (mergedomat[k] == number[k]) { // k is subset of merged
	  mergedmmat[k] = 2;
	  mergedcmat1[k] = g_dlarge;
	  mergedcmat2[k] = (mwi - within[k]) / (sum * (sum - 1) / 2 - number[k] * (number[k] - 1) / 2);
	  subclus.insert(k);
	} else {
	  mergedmmat[k] = 0;
	  mergedcmat1[k] = g_dlarge;
	  mergedcmat2[k] = g_dlarge;
	  if (mergedomat[k] == 0) {
	    nonol.insert(k);
	  }
	}
      }
      szsub = subclus.size();
      if (opt == 0) {
	for (si = subclus.begin(); si != subclus.end(); si++) {
	  si1 = si;
	  for (si1++; si1 != subclus.end(); si1++) {
	    if (*si > *si1) {
	      mmat[*si][*si1] = 2;
	    } else {
	      mmat[*si1][*si] = 2;
	    }
	  }
	}
      }
      if (g_debug >= 1) {
	cout << "subclus\t" << subclus.size() << endl;
	Mylib::PrintJoin(cout, subclus, " ");
	cout << endl;
	cout << "supclus\t" << supclus.size() << endl;
	Mylib::PrintJoin(cout, supclus, " ");
	cout << endl;
	cout << "nonol\t" << nonol.size() << endl;
	Mylib::PrintJoin(cout, nonol, " ");
	cout << endl;
      }
      // --------------------------------------------------
      //
      // update tumat, which is the only mat or vec that needs to be
      // updated besides the minor update to mmat we did earlier. Fill
      // mergedtumat and update mgtouse. Update touse, which should be
      // done even for flag != 1
      //
      // --------------------------------------------------
      mgtouse = -g_dlarge;
      for (si = subclus.begin(); si != subclus.end(); si++) {
	for (si1 = nonol.begin(); si1 != nonol.end(); si1++) {
	  // here we might get quadratic blowup but hopefully
	  // |subclus| * |nonol| wouldn't be large
	  // TODO: may remove the following to speedup after verifyting it's always true
	  assert(*si > *si1 && omat[*si][*si1] == 0 || *si < *si1 && omat[*si1][*si] == 0);
	  // if (*si > *si1 && omat[*si][*si1] == 0 || *si < *si1 && omat[*si1][*si] == 0) {
	  if (tumat[*si][*si1] > mergedcmat2[*si]) {
	    tumat[*si][*si1] = mergedcmat2[*si];
	    tupdate.insert(*si);
	    touse[*si] = -g_dlarge;
	  }
	  // }
	}
      }
      for (k = 0; k < curn; k++) {
	if (mergedomat[k] > 0) {
	  mergedtumat1[k] = -g_dlarge;
	  mergedtumat2[k] = -g_dlarge;
	  // don't need to upate touse[k] or mgtouse here
	} else {
	  mergedtumat1[k] = g_dlarge;
	  for (si = supclus.begin(); si != supclus.end(); si++) {
	    // here we might get quadratic blowup but hopefully
	    // |supclus| wouldn't be large
	    if (k > *si && omat[k][*si] == 0 || k < *si && omat[*si][k] == 0) {
	      if (mergedtumat1[k] > mergedcmat1[*si]) {
		mergedtumat1[k] = mergedcmat1[*si];
	      }
	    }
	  }
	  if (mgtouse < mergedtumat1[k] && mergedmmat[k] == 0) {
	    mgtouse = mergedtumat1[k];
	  }
	  mergedtumat2[k] = g_dlarge;
	  for (si = nonol.begin(); si != nonol.end(); si++) {	
	    // here we might get quadratic blowup but hopefully
	    // |nonol| wouldn't be large
	    if (cmat[k][*si] < g_dlarge) {
	      if (mergedtumat2[k] > cmat[k][*si]) {
		mergedtumat2[k] = cmat[k][*si];
	      }
	    }
	  }
	  if (touse[k] < mergedtumat2[k] && mergedmmat[k] == 0) {
	    touse[k] = mergedtumat2[k];
	  }
	}
      }
      if (g_debug >= 3) {
	cout << "mergedmat" << endl;
	Mylib::PrintJoin(cout, mergedmat, " ");
	cout << endl;
	cout << "mergedomat" << endl;
	Mylib::PrintJoin(cout, mergedomat, " ");
	cout << endl;
	cout << "mergedmmat" << endl;
	Mylib::PrintJoin(cout, mergedmmat, " ");
	cout << endl;
	cout << "mergedtumat1" << endl;
	Mylib::PrintJoin(cout, mergedtumat1, " ");
	cout << endl;
	cout << "mergedtumat2" << endl;
	Mylib::PrintJoin(cout, mergedtumat2, " ");
	cout << endl;
	cout << "mergedcmat1" << endl;
	Mylib::PrintJoin(cout, mergedcmat1, " ");
	cout << endl;
	cout << "mergedcmat2" << endl;
	Mylib::PrintJoin(cout, mergedcmat2, " ");
	cout << endl;
      }
    }
    if (g_debug >= 1) {
      cout << "tupdate\t" << tupdate.size() << endl;
      Mylib::PrintJoin(cout, tupdate, " ");
      cout << endl;
    }
    for (si = tupdate.begin(); si != tupdate.end(); si++) {
      for (k = 0; k < curn; k++) {
	if (touse[*si] < tumat[*si][k] && (k > *si && mmat[k][*si] == 0 || k < *si && mmat[*si][k] == 0)) {
	  touse[*si] = tumat[*si][k];
	}
      }
    }

    // --------------------------------------------------
    //
    // prune cluster that would make no difference. Do this regardless
    // of flag. Ensure pruning wouldn't affect anything else, which
    // should be identical as if no pruning. All mat's and vec's
    // except numsub and touse are only local so it should be
    // safer. numsub and esp. touse in result have global dependency
    // so care more
    //
    // --------------------------------------------------
    for (i = curn - 1; i >= 0; i--) {
      if (d - touse[i] >= ogap && numsub[i] == 0) {
	if (g_debug >= 1) {
	  cout << "remove cluster\t" << i << "\tfrom the pool\t" << result[clusterid[i]].size << endl;
	}
	for (k = 0; k < curn; k++) {
	  if (cmat[i][k] < g_dlarge) {
	    numsub[k]--;
	  }
	}
	if (mergedcmat2[i] < g_dlarge) {
	  szsub--;
	}
	result[clusterid[i]].upperb = d;
	result[clusterid[i]].lowerb = prevd;
	OutOtree(clusterid[i], result, dendout);
	curn--;
	if (i < curn) {
	  for (j = 0; j < i; j++) {
	    dmat[i][j] = dmat[curn][j];
	    omat[i][j] = omat[curn][j];
	    mmat[i][j] = mmat[curn][j];
	    tumat[i][j] = tumat[curn][j];
	    tumat[j][i] = tumat[j][curn];
	    cmat[i][j] = cmat[curn][j];
	    cmat[j][i] = cmat[j][curn];
	  }
	  tumat[i][i] = tumat[curn][curn];
	  cmat[i][i] = cmat[curn][curn];
	  for (j = i + 1; j <= curn; j++) {
	    dmat[j][i] = dmat[curn][j];
	    omat[j][i] = omat[curn][j];
	    mmat[j][i] = mmat[curn][j];
	    tumat[i][j] = tumat[curn][j];
	    tumat[j][i] = tumat[j][curn];
	    cmat[i][j] = cmat[curn][j];
	    cmat[j][i] = cmat[j][curn];
	  }
	  clusterid[i] = clusterid[curn];
	  number[i] = number[curn];
	  numsub[i] = numsub[curn];
	  within[i] = within[curn];
	  touse[i] = touse[curn];
	  if (flag == 1) {
	    mergedmat[i] = mergedmat[curn];
	    mergedomat[i] = mergedomat[curn];
	    mergedmmat[i] = mergedmmat[curn];
	    mergedtumat1[i] = mergedtumat1[curn];
	    mergedtumat2[i] = mergedtumat2[curn];
	    mergedcmat1[i] = mergedcmat1[curn];
	    mergedcmat2[i] = mergedcmat2[curn];
	  }
	}
      }
    }

    if (flag == 1) {
      if (curn < maxn) {
	k = curn;
	curn++;
      } else {
	min = g_dlarge;
	k = 0;
	for (i = 0; i < curn; i++) {
	  if (min > touse[i]) {
	    min = touse[i];
	    k = i;
	  }
	}
	cout << "max # clusters. Replace " << k << endl;
	for (i = 0; i < curn; i++) {
	  if (cmat[k][i] < g_dlarge) {
	    numsub[i]--;
	  }
	}
	if (mergedcmat2[k] < g_dlarge) {
	  szsub--;
	}
	result[clusterid[k]].upperb = d;
	result[clusterid[k]].lowerb = prevd;
	OutOtree(clusterid[k], result, dendout);
      }
      // --------------------------------------------------
      //
      // add the merged cluster to all mat's and vec's
      //
      // --------------------------------------------------
      for (i = 0; i < k; i++) {
	dmat[k][i] = mergedmat[i];
	omat[k][i] = mergedomat[i];
	mmat[k][i] = mergedmmat[i];
	tumat[k][i] = mergedtumat1[i];
	tumat[i][k] = mergedtumat2[i];
	cmat[k][i] = mergedcmat1[i];
	cmat[i][k] = mergedcmat2[i];
      }
      tumat[k][k] = -g_dlarge;
      cmat[k][k] = g_dlarge;
      for (i = k + 1; i < curn; i++) {
	dmat[i][k] = mergedmat[i];
	omat[i][k] = mergedomat[i];
	mmat[i][k] = mergedmmat[i];
	tumat[k][i] = mergedtumat1[i];
	tumat[i][k] = mergedtumat2[i];
	cmat[k][i] = mergedcmat1[i];
	cmat[i][k] = mergedcmat2[i];
      }
      clusterid[k] = result.size() - 1;
      number[k] = sum;
      within[k] = mwi;
      numsub[k] = szsub;
      touse[k] = mgtouse;
    }

    if (g_debug >= 2) {
      cout << "clusterid" << endl;
      Mylib::PrintJoin(cout, clusterid, " ");
      cout << endl; 
      cout << "number" << endl;
      Mylib::PrintJoin(cout, number, " ");
      cout << endl;
      cout << "within" << endl;
      Mylib::PrintJoin(cout, within, " ");
      cout << endl;
      cout << "numsub" << endl;
      Mylib::PrintJoin(cout, numsub, " ");
      cout << endl;
      cout << "touse" << endl;
      Mylib::PrintJoin(cout, touse, " ");
      cout << endl;
      PrintDistmatCluster(cout, dmat, curn, 1, result, clusterid);
      cout << "omat" << endl;
      PrintVecmat(cout, omat, curn);
      cout << "mmat" << endl;
      PrintVecmat(cout, mmat, curn);
      cout << "tumat" << endl;
      PrintVecmat(cout, tumat, curn);
      cout << "cmat" << endl;
      PrintVecmat(cout, cmat, curn);
      // PrintCluster(cout, curn, result, clusterid);
    }
    iter++;
  }
  if (iter >= g_maxiter) {
    cout << "Warning: stop at maxiter" << endl;
  }
  for (i = 0; i < curn; i++) {
    OutOtree(clusterid[i], result, dendout);
  }
}

/* ---------------------------------------------------------------------- */

// no overlap and no creation of identical cluster
double find_closest_pair_pmlotree(int n, double** dmat, int* ip, int* jp, const vector<intvec> & omat, const vector<intvec> & mmat)
/*
This function searches the distance matrix to find the pair with the shortest
distance between them. The indices of the pair are returned in ip and jp; the
distance itself is returned by the function.

n          (input) int
The number of elements in the distance matrix.

dmat (input) double**
A ragged array containing the distance matrix. The number of columns in each
row is the same as the row index.

ip         (output) int*
A pointer to the integer that is to receive the first index of the pair with
the shortest distance.

jp         (output) int*
A pointer to the integer that is to receive the second index of the pair with
the shortest distance.
*/
{
  int i, j;
  double temp;
  double distance = g_dlarge;
  for (i = 1; i < n; i++) {
    for (j = 0; j < i; j++) {
      if (omat[i][j] >= 0 && mmat[i][j] == 0) {
	temp = dmat[i][j];
	if (temp < distance) {
	  distance = temp;
	  *ip = i;
	  *jp = j;
	}
      }
    }
  }
  // we can also set mmat[*ip][*jp] = 1 here, which means if the pair
  //   is retrieved once, it won't be retrieved again. Right now, in
  //   the calling routine, it may be the pair is merged, the pair not
  //   merged because it creates an identical cluster, or the pair not
  //   merged because it exceeds ogap. In the first two cases, we set
  //   mmat, while in the last case, the entry in mmat would be throw
  //   away. So right now, it's equivalent to setting mmat = 1 here
  return distance;
}

// closest pair not excluded by omat or mmat, but might merge into a
// cluster that's identical to a previous one
double find_closest_pair_palotree(int n, double** dmat, int* ip, int* jp, const vector<intvec> & omat, vector<intvec> & mmat)
/*
This function searches the distance matrix to find the pair with the shortest
distance between them. The indices of the pair are returned in ip and jp; the
distance itself is returned by the function.

n          (input) int
The number of elements in the distance matrix.

dmat (input) double**
A ragged array containing the distance matrix. The number of columns in each
row is one less than the row index.

ip         (output) int*
A pointer to the integer that is to receive the first index of the pair with
the shortest distance.

jp         (output) int*
A pointer to the integer that is to receive the second index of the pair with
the shortest distance.
*/
{
  int i, j;
  double temp;
  double distance = g_dlarge;
  for (i = 1; i < n; i++) {
    for (j = 0; j < i; j++) {
      if (omat[i][j] == 0 && mmat[i][j] == 0) {
	temp = dmat[i][j];
	if (temp < distance) {
	  distance = temp;
	  *ip = i;
	  *jp = j;
	}
      }
    }
  }
  mmat[*ip][*jp] = 1;
  return distance;
}

int CalcOverlap(int oij, int oi, int oj, const intset & cij, const intset & ci, const intset & cj, const intset & c)
{
  int cnt, size;
  intset::const_iterator si;

  /*
  cout << "Manually compute ovelap ..." << endl;
  cnt = 0;
  for (si = c.begin(); si != c.end(); si++) {
    if (cij.find(*si) != cij.end()) {
      cnt++;
    }
  }
  // return (cnt < cij.size() && cnt < c.size()? 1: -1) * cnt;
  cnt = (cnt < cij.size() && cnt < c.size()? 1: -1) * cnt;
  */

  // check this is the same as brute force for all cases as in else {}
  if (oij <= 0 || oi <= 0 || oj <= 0) {
    if (oi < 0 && -oi == c.size()) {
      // assert(cnt == oi);
      return oi;
    }
    if (oj < 0 && -oj == c.size()) {
      // assert(cnt == oj);
      return oj;
    }
    if (oi == 0) {
      // assert(abs(oj) == cnt);
      return abs(oj);
    }
    if (oj == 0) {
      // assert(abs(oi) == cnt);
      return abs(oi);
    }
    if (oij < 0) {
      if (-oij == ci.size()) {
	// assert(cnt == oj);
	return oj;
      } else {
	// assert(cnt == oi);
	return oi;
      }
    } else { // oij > 0 so either oi < 0 or oj < 0 or oij == 0
      size = abs(oi) + abs(oj) - oij;
      // assert((size == c.size() || size == cij.size()? -1: 1) * size == cnt);
      return (size == c.size() || size == cij.size()? -1: 1) * size;
    }
  } else {
    // cout << "Manually compute ovelap ..." << endl;
    cnt = 0;
    for (si = c.begin(); si != c.end(); si++) {
      if (cij.find(*si) != cij.end()) {
	cnt++;
      }
    }
    return (cnt < cij.size() && cnt < c.size()? 1: -1) * cnt;
  }
}

/*
void palotree(int nele, int maxn, double ogap, double** distmatrix, vector<Onode> & result)

Purpose
=======

The pmlotree routine performs clustering using pairwise average
linkage on the given distance matrix while allow overlapping

Arguments
=========

nele: # elements to be clustered.

distmatrix: the distance matrix, with nele rows, each row being filled
up to the diagonal. The elements on the diagonal do not exist and they
are assumed to be zero. The distance matrix will be modified by this
routine.

result: touse to return vector of Onode structs, describing the
clustering solution consisting of a set of nodes. See src/cluster.h for
a description of the Onode structure.

If a memory error occurs, pmlotree returns NULL???
==================================================================

{
  // curn: # potential merging candidates
  int i, j, curn = nele, is, js, k, flag;
  double d, min, sum, di, dj;
  Onode curnode;
  // clusterid: each candidate in distmatrix is a cluster (subtree).
  //  Assign it an ID. The singletons will have the ID equal to their
  //  original position in distmatrix. The cluster created through
  //  merging is assigned an ID, which is one less the negative of its
  //  postion in result
  intvec clusterid(maxn), number(maxn), mergedomat(maxn), mergedmmat(maxn);
  // info about 'touse' is also in results except for the singletons,
  //   which can be not equal to g_dlarge. May remove 'touse' if we can
  //   take care of that
  dblvec mergedist(maxn), mergedtouse1(maxn), mergedtouse2(maxn);
  // mmat: 1: merged. 2: if merge, creat identical cluster to someone
  // within ogap so merge skipped. 0: otherwise
  vector<intvec> omat(maxn), mmat(maxn);
  vector<dblvec> touse(maxn);
  // subclus: all clusters that's a subset of the to-be-merged cluster
  intset newclus;
  intset::const_iterator si;

  result.clear();
  for (i = 0; i < maxn; i++) {
    touse[i].resize(maxn);
    omat[i].resize(i);
    mmat[i].resize(i);
    for (j = 0; j < i; j++) {
      touse[i][j] = g_dlarge;
      omat[i][j] = 0;
      mmat[i][j] = 0;
    }
    touse[i][i] = -g_dlarge;
    for (j = i + 1; j < maxn; j++) {
      touse[i][j] = g_dlarge;
    }
  }
  // setup a list specifying to which cluster each singleton belongs,
  // & keep track of # elements in each cluster (needed to calculate
  // the average)
  for (i = 0; i < nele; i++) {
    number[i] = 1;
    clusterid[i] = i;
    curnode.left = -1;
    curnode.right = -1;
    curnode.distance = 0;
    curnode.upperb = g_dlarge;
    curnode.membs.clear();
    curnode.membs.insert(i);
    result.push_back(curnode);
  }

  while (curn > 1) {
    cout << "new iter with " << curn << " clusters" << endl;
    is = 1;
    js = 0;
    // closest pair not excluded by omat or mmat, but might merge into
    // a cluster that's identical to a previous one
    //
    // TODO: check if d is non-decreasing
    d = find_closest_pair_palotree(curn, distmatrix, &is, &js, omat, mmat);
    if (d == g_dlarge) {
      cout << "can't find any close pair" << endl;
      break;
    }
    Onode & nis = result[clusterid[is]];
    Onode & njs = result[clusterid[js]];
    if (g_debug >= 1) {
      cout << "closest pair:\t" << is << '\t' << js << endl;
      Mylib::PrintJoin(cout, nis.membs, " ");
      cout << endl;
      Mylib::PrintJoin(cout, njs.membs, " ");
      cout << endl << "distance: " << d << endl;
    }
    flag = 0;
    if (d - touse[is][js] < ogap && d - touse[js][is] < ogap) {
      const intset & iset = nis.membs;
      const intset & jset = njs.membs;
      newclus = iset;
      for (si = jset.begin(); si != jset.end(); si++) {
	newclus.insert(*si);
      }
      if (g_debug >= 1) {
	cout << "newclus\t";
	Mylib::PrintJoin(cout, newclus, " ");
	cout << endl;
      }

      sum = number[is] + number[js]; // since no overlap between is and js
      for (k = 0; k < curn; k++) {
	if (k == is || k == js) {
	  continue;
	}
	Onode & nk = result[clusterid[k]];
	const intset & ket = nk.membs;
	i = (is > k? omat[is][k]: omat[k][is]);
	j = (js > k? omat[js][k]: omat[k][js]);
	di = (is > k? distmatrix[is][k]: distmatrix[k][is]);
	dj = (js > k? distmatrix[js][k]: distmatrix[k][js]);
	mergedomat[k] = i + j;
	if (mergedomat[k] == newclus.size() && mergedomat[k] == ket.size()) {
	  cout << is << ' ' << js << " merge to " << k << endl;
	  flag = -k - 1;
	  break;
	}
	mergedmmat[k] = 0;
	if (mergedomat[k] == 0) {
	  mergedist[k] = (di * number[is] + dj * number[js]) / sum;
	}
      }
      mergedomat[is] = iset.size();
      mergedmmat[is] = 2;
      mergedomat[js] = jset.size();
      mergedmmat[js] = 2;
      // do following regardless of flag
      for (k = 0; k < curn; k++) {
	if (k == is || k == js) {
	  continue;
	}
	if (flag < 0) {
	  i = -flag - 1;
	  if ((i > k? omat[i][k]: omat[k][i]) > 0) {
	    // k == i excluded
	    continue;
	  }
	} else {
	  if (mergedomat[k] > 0) {
	    continue;
	  }
	}
	// is or js not overlap with k b/c their superset don't
	if (touse[is][k] > d) {
	  touse[is][k] = d;
	}
	if (touse[js][k] > d) {
	  touse[js][k] = d;
	}
      }
      touse[is][js] = -g_dlarge;
      touse[js][is] = -g_dlarge;
      nis.upperb = d;
      njs.upperb = d;
      for (k = 0; k < curn; k++) {
	if (nis.upperb < touse[is][k]) {
	  nis.upperb = touse[is][k];
	}
	if (njs.upperb < touse[js][k]) {
	  njs.upperb = touse[js][k];
	}
      }
      if (flag < 0) {
	// we should probably also update the distance for the cluster
	// -flag - 1 to take the min or max
	mmat[is][js] = 2;
      } else {
	flag = 1;
	mmat[is][js] = 1;
	curnode.distance = d;
	curnode.upperb = -g_dlarge;
	curnode.left = clusterid[is];
	curnode.right = clusterid[js];
	curnode.membs = newclus;
	for (k = 0; k < curn; k++) {
	  if (mergedomat[k] == 0) {
	    assert(k != is && k != js);
	    mergedtouse1[k] = g_dlarge;
	    curnode.upperb = g_dlarge;
	    // TODO: not exactly correct. Need to keep another mat to
	    // compute correctly
	    mergedtouse2[k] = (touse[k][is] > touse[k][js]? touse[k][is]: touse[k][js]);
	  } else {
	    mergedtouse1[k] = -g_dlarge;
	    mergedtouse2[k] = -g_dlarge;
	  }
	}
	result.push_back(curnode);
	if (g_debug >= 3) {
	  cout << "mergedomat" << endl;
	  Mylib::PrintJoin(cout, mergedomat, " ");
	  cout << endl;
	  cout << "mergedmmat" << endl;
	  Mylib::PrintJoin(cout, mergedmmat, " ");
	  cout << endl;
	  cout << "mergedtouse1" << endl;
	  Mylib::PrintJoin(cout, mergedtouse1, " ");
	  cout << endl;
	  cout << "mergedtouse2" << endl;
	  Mylib::PrintJoin(cout, mergedtouse2, " ");
	  cout << endl;
	  cout << "mergedist" << endl;
	  Mylib::PrintJoin(cout, mergedist, " ");
	  cout << endl;
	}
      }
    }
    if (g_debug >= 1) {
      cout << "flag\t" << flag << endl;
    }
    for (i = curn - 1; i >= 0; i--) {
      if (d - result[clusterid[i]].upperb >= ogap) {
	if (g_debug >= 1) {
	  cout << "remove cluster " << i << " from the pool:" << endl;
	  Mylib::PrintJoin(cout, result[clusterid[i]].membs, " ");
	  cout << endl;
	}
	curn--;
	if (i < curn) {
	  for (j = 0; j < i; j++) {
	    distmatrix[i][j] = distmatrix[curn][j];
	    omat[i][j] = omat[curn][j];
	    mmat[i][j] = mmat[curn][j];
	    touse[i][j] = touse[curn][j];
	    touse[j][i] = touse[j][curn];
	  }
	  touse[i][i] = touse[curn][curn];
	  for (j = i + 1; j <= curn; j++) {
	    distmatrix[j][i] = distmatrix[curn][j];
	    omat[j][i] = omat[curn][j];
	    mmat[j][i] = mmat[curn][j];
	    touse[i][j] = touse[curn][j];
	    touse[j][i] = touse[j][curn];
	  }
	  clusterid[i] = clusterid[curn];
	  number[i] = number[curn];
	  mergedist[i] = mergedist[curn];
	  mergedomat[i] = mergedomat[curn];
	  mergedmmat[i] = mergedmmat[curn];
	  mergedtouse1[i] = mergedtouse1[curn];
	  mergedtouse2[i] = mergedtouse2[curn];
	}
      }
    }
    // PrintVecmat(cout, omat, curn);
    // PrintVecmat(cout, mmat, curn);

    if (flag == 1) {
      if (curn < maxn) {
	k = curn;
	curn++;
      } else {
	min = g_dlarge;
	k = 0;
	for (i = 0; i < curn; i++) {
	  if (min > result[clusterid[i]].upperb) {
	    min = result[clusterid[i]].upperb;
	    k = i;
	  }
	}
      }
      for (i = 0; i < k; i++) {
	distmatrix[k][i] = mergedist[i];
	omat[k][i] = mergedomat[i];
	mmat[k][i] = mergedmmat[i];
	touse[k][i] = mergedtouse1[i];
	touse[i][k] = mergedtouse2[i];
      }
      touse[k][k] = -g_dlarge;
      for (i = k + 1; i < curn; i++) {
	distmatrix[i][k] = mergedist[i];
	omat[i][k] = mergedomat[i];
	mmat[i][k] = mergedmmat[i];
	touse[k][i] = mergedtouse1[i];
	touse[i][k] = mergedtouse2[i];
      }
      clusterid[k] = result.size() - 1;
      number[k] = sum;
    }

    if (g_debug >= 2) {
      cout << "clusterid" << endl;
      Mylib::PrintJoin(cout, clusterid, " ");
      cout << endl; 
      cout << "number" << endl;
      Mylib::PrintJoin(cout, number, " ");
      cout << endl;
      cout << "omat" << endl;
      PrintVecmat(cout, omat, curn);
      cout << "mmat" << endl;
      PrintVecmat(cout, mmat, curn);
      cout << "touse" << endl;
      PrintVecmat(cout, touse, curn);
      PrintDistmatCluster(cout, distmatrix, curn, 1, result, clusterid);
      // PrintCluster(cout, curn, result, clusterid);
    }
  }
}
*/

// The palotree routine performs clustering using pairwise average
// linkage on the given distance matrix while allow overlapping
//
// nele: # elements to be clustered.
//
// maxn: max # candidate clusters ever allowed in pool
//
// dmat: the distance matrix, with nele rows, each row being filled up
// to the diagonal. The elements on the diagonal do not exist. The
// distance matrix will be modified by this routine.
//
// result: used to return vector of Onode structs, describing the
// clusters ever appeared. See util.h for a description of the Onode
// structure.
//
// dendout: output file for the dendogram
void palotree(int nele, int maxn, double ogap, double** dmat, vector<Onode> & result, ostream & dendout)
{
  // curn: # potential merging candidates
  int i, j, curn = nele, is, js, k, flag;
  double prevd, d = -g_dlarge, min, sum, mwi, mgtouse, di, dj;
  Onode curnode;
  // clusterid: each candidate in dmat is a cluster (subtree).  Assign
  // it an ID to be the its position in result
  intvec clusterid(maxn), number(maxn);
  dblvec within(maxn), touse(maxn);
  // mmat: 1: merged. 2: if merge, creat identical cluster to someone
  // within ogap so merge skipped. 0: otherwise
  vector<intvec> omat(maxn), mmat(maxn);
  vector<dblvec> cmat(maxn); // dmat
  intvec mergedomat(maxn), mergedmmat(maxn);
  // cmat[merge][i] = mergedcmat1[i]. cmat[i][merge] = mergedcmat2[i]
  dblvec mergedmat(maxn), mergedcmat1(maxn), mergedcmat2(maxn);
  // subclus: all clusters that's a subset of the to-be-merged cluster
  intset newclus, subclus;
  intset::const_iterator si, si1;

  // --------------------------------------------------
  //
  // initialize mat's, vec's, and result for singletons
  //
  // --------------------------------------------------
  result.clear();
  for (i = 0; i < maxn; i++) {
    omat[i].resize(i);
    mmat[i].resize(i);
    cmat[i].resize(maxn);
    for (j = 0; j < i; j++) {
      omat[i][j] = 0;
      mmat[i][j] = 0;
      cmat[i][j] = g_dlarge;
    }
    cmat[i][j] = g_dlarge;
    for (j = i + 1; j < maxn; j++) {
      cmat[i][j] = g_dlarge;
    }
  }
  for (i = 0; i < nele; i++) {
    clusterid[i] = i;
    number[i] = 1;
    within[i] = 0;
    touse[i] = g_dlarge;
    curnode.left = -1;
    curnode.right = -1;
    curnode.distance = -g_dlarge;
    curnode.upperb = g_dlarge;
    curnode.lowerb = g_dlarge;
    curnode.touse = g_dlarge;
    curnode.membs.clear();
    curnode.membs.insert(i);
    result.push_back(curnode);
  }

  while (curn > 1) {
    is = 1;
    js = 0;
    // TODO: check if d is non-decreasing
    prevd = d;
    d = find_closest_pair_palotree(curn, dmat, &is, &js, omat, mmat);
    cout << "new iter. # cluster\t" << curn << "\tdistance\t" << d << endl;
    if (d == g_dlarge) {
      cout << "can't find any close pair" << endl;
      break;
    }
    Onode & nis = result[clusterid[is]];
    Onode & njs = result[clusterid[js]];
    if (g_debug >= 1) {
      cout << "closest pair:\t" << is << '\t' << js << endl;
      Mylib::PrintJoin(cout, nis.membs, " ");
      cout << endl;
      Mylib::PrintJoin(cout, njs.membs, " ");
      cout << endl;
    }
    // --------------------------------------------------
    //
    // flag status of the pair found: merge to existing cluster (< 0),
    // used (0), or okay to merge (1). Fill mergedomat along the way
    // because it's needed for checking the status
    //
    // --------------------------------------------------
    flag = 0;
    if (d - touse[is] < ogap && d - touse[js] < ogap) {
      const intset & iset = nis.membs;
      const intset & jset = njs.membs;
      newclus = iset;
      for (si = jset.begin(); si != jset.end(); si++) {
	newclus.insert(*si);
      }
      if (g_debug >= 1) {
	cout << "newclus\t";
	Mylib::PrintJoin(cout, newclus, " ");
	cout << endl;
      }

      for (k = 0; k < curn; k++) {
	if (k == is || k == js) {
	  continue;
	}
	// Onode & nk = result[clusterid[k]]; // only for excluding exact match
	// const intset & ket = nk.membs; // only for excluding exact match
	i = (is > k? omat[is][k]: omat[k][is]);
	j = (js > k? omat[js][k]: omat[k][js]);
	mergedomat[k] = i + j;
	// only for excluding exact match
	// if (mergedomat[k] == newclus.size() && mergedomat[k] == ket.size()) {
	// only for excluding subset. Can remove the following if it's never true
	if (mergedomat[k] == newclus.size()) {
	  // TODO: okay if the merged set is a subset. May change
	  cout << is << ' ' << js << " merge to a subset of " << k << endl;
	  flag = -k - 1;
	  break;
	}
      }
      if (flag == 0) {
	flag = 1;
	mergedomat[is] = iset.size();
	mergedomat[js] = jset.size();
      }
    }

    if (g_debug >= 1) {
      cout << "result\t" << result.size() << "\tflag\t" << flag << endl;
    }
    if (flag == 0) {
      // this status should never be kept long because either is or js
      // would be removed
      mmat[is][js] = 3;
    } else if (flag < 0) {
      mmat[is][js] = 2;
    } else {
      // --------------------------------------------------
      //
      // If okay to merge, fill mergedmat. Minor update to mmat
      //
      // --------------------------------------------------
      curnode.distance = d;
      curnode.upperb = g_dlarge; // will be updated later
      curnode.lowerb = g_dlarge; // will be updated later
      curnode.touse = g_dlarge; // will be updated later
      curnode.left = clusterid[is];
      curnode.right = clusterid[js];
      curnode.membs = newclus;
      result.push_back(curnode);
      mmat[is][js] = 1;
      sum = number[is] + number[js]; // b/c no overlap between is and js
      mwi = within[is] + within[js] + d * number[is] * number[js];
      for (k = 0; k < curn; k++) {
	if (k == is || k == js) {
	  continue;
	}
	di = (is > k? dmat[is][k]: dmat[k][is]);
	dj = (js > k? dmat[js][k]: dmat[k][js]);
	if (mergedomat[k] == 0) {
	  mergedmat[k] = (di * number[is] + dj * number[js]) / sum;
	}
      }
      // --------------------------------------------------
      //
      // Fill mergemmat and mergedcmat. Update mgtouse. Udate touse,
      // which is the only mat or vec that needs to be updated besides
      // the update to mmat we did earlier and here
      //
      // --------------------------------------------------
      subclus.clear();
      mgtouse = g_dlarge;
      for (k = 0; k < curn; k++) {
	Onode & nk = result[clusterid[k]];
	const intset & ket = nk.membs;
	if (mergedomat[k] == newclus.size()) { // merged is subset of k
	  // TODO: can remove this branch if we exclude subset
	  assert(false);
	  mergedmmat[k] = 2;
	  mergedcmat1[k] = (within[k] - mwi) / (number[k] * (number[k] - 1) / 2 - sum * (sum - 1) / 2);
	  mergedcmat2[k] = g_dlarge;
	  if (mgtouse > mergedcmat1[k]) {
	    mgtouse = mergedcmat1[k];
	  }
	} else if (mergedomat[k] == ket.size()) { // k is subset of merged
	  mergedmmat[k] = 2;
	  mergedcmat1[k] = g_dlarge;
	  mergedcmat2[k] = (mwi - within[k]) / (sum * (sum - 1) / 2 - number[k] * (number[k] - 1) / 2);
	  if (touse[k] > mergedcmat2[k]) {
	    touse[k] = mergedcmat2[k];
	  }
	  subclus.insert(k);
	} else {
	  mergedmmat[k] = 0;
	  mergedcmat1[k] = g_dlarge;
	  mergedcmat2[k] = g_dlarge;
	}
      }
      // only for excluding subset
      for (si = subclus.begin(); si != subclus.end(); si++) {
	si1 = si;
	for (si1++; si1 != subclus.end(); si1++) {
	  if (*si > *si1) {
	    mmat[*si][*si1] = 2;
	  } else {
	    mmat[*si1][*si] = 2;
	  }
	}
      }
      if (g_debug >= 3) {
	cout << "subclus\t" << subclus.size() << endl;
	Mylib::PrintJoin(cout, subclus, " ");
	cout << endl;
	cout << "mergedmat" << endl;
	Mylib::PrintJoin(cout, mergedmat, " ");
	cout << endl;
	cout << "mergedomat" << endl;
	Mylib::PrintJoin(cout, mergedomat, " ");
	cout << endl;
	cout << "mergedmmat" << endl;
	Mylib::PrintJoin(cout, mergedmmat, " ");
	cout << endl;
	cout << "mergedcmat1" << endl;
	Mylib::PrintJoin(cout, mergedcmat1, " ");
	cout << endl;
	cout << "mergedcmat2" << endl;
	Mylib::PrintJoin(cout, mergedcmat2, " ");
	cout << endl;
      }
    }
    // --------------------------------------------------
    //
    // prune cluster that would make no difference. Do this regardless
    // of flag. Ensure pruning wouldn't affect anything else, which
    // should be identical as if no pruning. All mat's and vec's
    // except touse are only local so it should be safer. touse in
    // result have global dependency so care more
    //
    // --------------------------------------------------
    for (i = curn - 1; i >= 0; i--) {
      if (d - touse[i] >= ogap) {
	if (g_debug >= 1) {
	  cout << "remove cluster " << i << " from the pool:" << endl;
	  Mylib::PrintJoin(cout, result[clusterid[i]].membs, " ");
	  cout << endl;
	}
	result[clusterid[i]].upperb = d;
	result[clusterid[i]].lowerb = prevd;
	result[clusterid[i]].touse = touse[i];
	OutOtree(clusterid[i], result, dendout);
	curn--;
	if (i < curn) {
	  for (j = 0; j < i; j++) {
	    dmat[i][j] = dmat[curn][j];
	    omat[i][j] = omat[curn][j];
	    mmat[i][j] = mmat[curn][j];
	    cmat[i][j] = cmat[curn][j];
	    cmat[j][i] = cmat[j][curn];
	  }
	  cmat[i][i] = cmat[curn][curn];
	  for (j = i + 1; j <= curn; j++) {
	    dmat[j][i] = dmat[curn][j];
	    omat[j][i] = omat[curn][j];
	    mmat[j][i] = mmat[curn][j];
	    cmat[i][j] = cmat[curn][j];
	    cmat[j][i] = cmat[j][curn];
	  }
	  clusterid[i] = clusterid[curn];
	  number[i] = number[curn];
	  within[i] = within[curn];
	  touse[i] = touse[curn];
	  if (flag == 1) {
	    mergedmat[i] = mergedmat[curn];
	    mergedomat[i] = mergedomat[curn];
	    mergedmmat[i] = mergedmmat[curn];
	    mergedcmat1[i] = mergedcmat1[curn];
	    mergedcmat2[i] = mergedcmat2[curn];
	  }
	}
      }
    }

    if (flag == 1) {
      if (curn < maxn) {
	k = curn;
	curn++;
      } else {
	min = g_dlarge;
	k = 0;
	for (i = 0; i < curn; i++) {
	  if (min > touse[i]) {
	    min = touse[i];
	    k = i;
	  }
	}
	cout << "max # clusters. Replace\t" << k << "\twith touse\t" << min << endl;
	result[clusterid[k]].upperb = d;
	result[clusterid[k]].lowerb = prevd;
	result[clusterid[k]].touse = touse[k];
	OutOtree(clusterid[k], result, dendout);
      }
      // --------------------------------------------------
      //
      // add the merged cluster to all mat's and vec's
      //
      // --------------------------------------------------
      for (i = 0; i < k; i++) {
	dmat[k][i] = mergedmat[i];
	omat[k][i] = mergedomat[i];
	mmat[k][i] = mergedmmat[i];
	cmat[k][i] = mergedcmat1[i];
	cmat[i][k] = mergedcmat2[i];
      }
      cmat[k][k] = g_dlarge;
      for (i = k + 1; i < curn; i++) {
	dmat[i][k] = mergedmat[i];
	omat[i][k] = mergedomat[i];
	mmat[i][k] = mergedmmat[i];
	cmat[k][i] = mergedcmat1[i];
	cmat[i][k] = mergedcmat2[i];
      }
      clusterid[k] = result.size() - 1;
      number[k] = sum;
      within[k] = mwi;
      touse[k] = mgtouse;
    }

    if (g_debug >= 2) {
      cout << "clusterid" << endl;
      Mylib::PrintJoin(cout, clusterid, " ");
      cout << endl; 
      cout << "number" << endl;
      Mylib::PrintJoin(cout, number, " ");
      cout << endl;
      cout << "within" << endl;
      Mylib::PrintJoin(cout, within, " ");
      cout << endl;
      cout << "touse" << endl;
      Mylib::PrintJoin(cout, touse, " ");
      cout << endl;
      PrintDistmatCluster(cout, dmat, curn, 1, result, clusterid);
      cout << "omat" << endl;
      PrintVecmat(cout, omat, curn);
      cout << "mmat" << endl;
      PrintVecmat(cout, mmat, curn);
      cout << "cmat" << endl;
      PrintVecmat(cout, cmat, curn);
      // PrintCluster(cout, curn, result, clusterid);
    }
  }
  for (i = 0; i < curn; i++) {
    OutOtree(clusterid[i], result, dendout);
  }
}

void AddClus(int id, const vector<Onode> & otree, intset & membs)
{
  if (otree[id].left == -1) {
    membs.insert(id);
  } else {
    AddClus(otree[id].left, otree, membs);
  }
  if (otree[id].right == -1) {
    membs.insert(id);
  } else {
    AddClus(otree[id].right, otree, membs);
  }
}
